using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Runtime.InteropServices;

using Tao.OpenGl;
using Tao.Platform.Windows;

namespace BezierSurface
{
    public partial class bezierSurfaceForm : Form
    {
        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        extern static bool DestroyIcon(IntPtr handle);

        #region Attributes
        Icon newIcon;
        private int viewportW = 800;
        private int viewportH = 600;
        /// <summary>Half width of window</summary>
        private double winX = 4.0;
        /// <summary>Half height of window</summary>
        private double winY = 4.0;
        int drawFlag = 0;
        int rotateFlag = 0;
        private Rectangle rec;

        #region Control Points
        /// <summary>four sets each of four control points </summary>
        private double[, ,] ctrlPts ={ { { -3.0, -1.5, 3.0 },//-2.0, -1.5, 2.1
                                        { -1.0, 2.5, 3.0 },
                                        { 1.0, 2.5, 3.0 },
                                        { 3.0, -1.5, 3.0 } },//2.0, -1.5, 2.1

                                      { { -2.0, -1.5, 1.0 },
                                        { -1.0, 1.5, 1.0 },
                                        { 1.0, 1.5, 1.0 },
                                        { 2.0, -1.5, 1.0} },

                                      { { -2.0, -1.5, -1.0 },
                                        { -1.0, 1.5, -1.0 },
                                        { 1.0, 1.5, -1.0  },
                                        { 2.0, -1.5, -1.0 } },

                                      { { -3.0, -1.5, -3.0 },//-2.0, -1.5, -2.1
                                        { -1.0, 2.5, -3.0  },
                                        { 1.0, 2.5, -3.0 },
                                        { 3.0, -1.5, -3.0 } } };//2.0, -1.5, -2.1
        #endregion

        /// <summary>number of vertices as curve points along v</summary>
        private static int NUM1 = 20;
        /// <summary>number of vertices as curve points along u</summary>
        private static int NUM2 = 20;
        private double[, ,] curvePts;
        private double[, ,] normals;
        /// <summary>Angle for rotation</summary>
        private static int theta=0;

        //Light values
        private float[] lightPos ={ -1.5f, 1.5f, 1.5f, 0.0f };
        private float[] amb ={ 0.3f, 0.3f, 0.3f, 1.0f };
        private float[] modelAmbient ={ 0.3f, 0.3f, 0.3f, 1.0f };
        private float[] diffuse ={ 0.9f, 0.0f, 0.0f, 1.0f };
        private float[] specular ={ 0.6f, 0.6f, 0.6f, 1.0f };
                
        #endregion
        #region Methods
        public bezierSurfaceForm()
        {
            InitializeComponent();
            this.SetBounds(0, 0, viewportW, viewportH);
            glControl.InitializeContexts();
            initializeOpenGl();
            curvePts = new double[NUM1 + 1, NUM2 + 1, 3];
            normals = new double[NUM1 + 1, NUM2 + 1, 3];
            this.generateCurvePts();
            this.generateNormals();  
        }
        public void getIconFromBitmap()
        {
            Bitmap myBitmap = new Bitmap(@"..\..\Images\rose12.jpg");
            IntPtr Hicon = myBitmap.GetHicon();
            newIcon = Icon.FromHandle(Hicon);
            this.Icon = newIcon;
            //DestroyIcon(newIcon.Handle);
        }
        private void glControl_Paint(object sender, PaintEventArgs e)
        {
            Gl.glClear(Gl.GL_COLOR_BUFFER_BIT | Gl.GL_DEPTH_BUFFER_BIT);
            if (rotateFlag == 1)
            {
                theta++;
            }
                Gl.glPushMatrix();
                    Gl.glRotated(theta, 0, 1, 0);
                    if (drawFlag == 1)
                        drawBezierSurfaceMesh();
                    if (drawFlag == 2)
                        drawBezierSurface();               
                Gl.glPopMatrix();
              
                //Thread.Sleep(200);
                glControl.Invalidate();                
        }
        private void glControl_Resize(object sender, EventArgs e)
        {
            rec = ClientRectangle;
            viewportW = rec.Width;
            viewportH = rec.Height;

            Gl.glViewport(0, 0, viewportW, viewportH);
            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glLoadIdentity();
            if (viewportW > viewportH)
                Gl.glOrtho(-winX * viewportW / viewportH, winX * viewportW / viewportH, -winY, winY, -10, 10);
            else
                Gl.glOrtho(-winX, winX, winY * viewportH / viewportW, -winY * viewportH / viewportW, -10, 10);

            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glLoadIdentity();
            Glu.gluLookAt(1.0, 1.0, 2.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

            glControl.Invalidate();
        }

        public void initializeOpenGl()
        {

            Gl.glViewport(0, 0, viewportW, viewportH);
            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glLoadIdentity();

            Gl.glOrtho(-winX * (8.0 / 6.0), winX * (8.0 / 6.0), -winY, winY, -10, 10);

            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glLoadIdentity();
            Glu.gluLookAt(1.0, 1.0, 2.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

            Gl.glClearColor(0, 0, 0, 1);
            Gl.glClear(Gl.GL_COLOR_BUFFER_BIT | Gl.GL_DEPTH_BUFFER_BIT);
            Gl.glEnable(Gl.GL_DEPTH_TEST);
            Gl.glShadeModel(Gl.GL_SMOOTH);
            initLights();

        }
        public void initLights()
        {
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_POSITION, lightPos);
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_AMBIENT, amb);
            //Gl.glLightModelfv(Gl.GL_LIGHT_MODEL_AMBIENT, modelAmbient);
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_DIFFUSE, diffuse);
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_SPECULAR, specular);

            Gl.glEnable(Gl.GL_LIGHTING);
            Gl.glEnable(Gl.GL_LIGHT0);
        }   
        public void drawBezierSurfaceMesh()
        {
            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glColor3d(0.245, 8.765, 0.1543);
          
            for(int m=0;m<NUM1;m++)
            {
                Gl.glBegin(Gl.GL_LINE_STRIP);
                for (int n = 0; n < NUM2;n++ )
                {
                    Gl.glVertex3d(curvePts[m, n, 0], curvePts[m, n, 1], curvePts[m, n, 2]);
                }
                Gl.glEnd();
            }
            for (int n = 0; n < NUM1; n++)
            {
                Gl.glBegin(Gl.GL_LINE_STRIP);
                for (int m = 0; m < NUM2; m++)
                {
                    Gl.glVertex3d(curvePts[m, n, 0], curvePts[m, n, 1], curvePts[m, n, 2]);
                }
                Gl.glEnd();
            }
            
            drawBezier3DHull();
        }
        public void drawBezierSurface()
        {
            Gl.glEnable(Gl.GL_LIGHTING);
            for (int m = 0; m < NUM1; m++)
            {
                for (int n = 0; n < NUM2; n++)
                {
                    Gl.glBegin(Gl.GL_QUADS);
                        Gl.glNormal3d(normals[m, n, 0], normals[m, n, 1], normals[m, n, 2]);
                        Gl.glVertex3d(curvePts[m, n, 0], curvePts[m, n, 1], curvePts[m, n, 2]);

                        Gl.glNormal3d(normals[m, n + 1, 0], normals[m, n + 1, 1], normals[m, n + 1, 2]);
                        Gl.glVertex3d(curvePts[m, n + 1, 0], curvePts[m, n + 1, 1], curvePts[m, n + 1, 2]);

                        Gl.glNormal3d(normals[m + 1, n + 1, 0], normals[m + 1, n + 1, 1], normals[m + 1, n + 1, 2]);
                        Gl.glVertex3d(curvePts[m + 1, n + 1, 0], curvePts[m + 1, n + 1, 1], curvePts[m + 1, n + 1, 2]);

                        Gl.glNormal3d(normals[m + 1, n, 0], normals[m + 1, n, 1], normals[m + 1, n, 2]);
                        Gl.glVertex3d(curvePts[m + 1, n, 0], curvePts[m + 1, n, 1], curvePts[m + 1, n, 2]);
                    Gl.glEnd();
                }
            }
            //drawBezier3DHull();
        }
        public void drawBezier3DHull()
        {
            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glColor3d(0.9, 0.956, 0.92);

            Gl.glBegin(Gl.GL_LINE_STRIP);
            Gl.glVertex3d(ctrlPts[0, 0, 0], ctrlPts[0, 0, 1], ctrlPts[0, 0, 2]);
            Gl.glVertex3d(ctrlPts[0, 1, 0], ctrlPts[0, 1, 1], ctrlPts[0, 1, 2]);
            Gl.glVertex3d(ctrlPts[0, 2, 0], ctrlPts[0, 2, 1], ctrlPts[0, 2, 2]);
            Gl.glVertex3d(ctrlPts[0, 3, 0], ctrlPts[0, 3, 1], ctrlPts[0, 3, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
            Gl.glVertex3d(ctrlPts[1, 0, 0], ctrlPts[1, 0, 1], ctrlPts[1, 0, 2]);
            Gl.glVertex3d(ctrlPts[1, 1, 0], ctrlPts[1, 1, 1], ctrlPts[1, 1, 2]);
            Gl.glVertex3d(ctrlPts[1, 2, 0], ctrlPts[1, 2, 1], ctrlPts[1, 2, 2]);
            Gl.glVertex3d(ctrlPts[1, 3, 0], ctrlPts[1, 3, 1], ctrlPts[1, 3, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
            Gl.glVertex3d(ctrlPts[2, 0, 0], ctrlPts[2, 0, 1], ctrlPts[2, 0, 2]);
            Gl.glVertex3d(ctrlPts[2, 1, 0], ctrlPts[2, 1, 1], ctrlPts[2, 1, 2]);
            Gl.glVertex3d(ctrlPts[2, 2, 0], ctrlPts[2, 2, 1], ctrlPts[2, 2, 2]);
            Gl.glVertex3d(ctrlPts[2, 3, 0], ctrlPts[2, 3, 1], ctrlPts[2, 3, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
            Gl.glVertex3d(ctrlPts[3, 0, 0], ctrlPts[3, 0, 1], ctrlPts[3, 0, 2]);
            Gl.glVertex3d(ctrlPts[3, 1, 0], ctrlPts[3, 1, 1], ctrlPts[3, 1, 2]);
            Gl.glVertex3d(ctrlPts[3, 2, 0], ctrlPts[3, 2, 1], ctrlPts[3, 2, 2]);
            Gl.glVertex3d(ctrlPts[3, 3, 0], ctrlPts[3, 3, 1], ctrlPts[3, 3, 2]);
            Gl.glEnd();

            //lines of different set of control points           
            Gl.glBegin(Gl.GL_LINE_STRIP);
                Gl.glVertex3d(ctrlPts[0, 0, 0], ctrlPts[0, 0, 1], ctrlPts[0, 0, 2]);
                Gl.glVertex3d(ctrlPts[1, 0, 0], ctrlPts[1, 0, 1], ctrlPts[1, 0, 2]);
                Gl.glVertex3d(ctrlPts[2, 0, 0], ctrlPts[2, 0, 1], ctrlPts[2, 0, 2]);
                Gl.glVertex3d(ctrlPts[3, 0, 0], ctrlPts[3, 0, 1], ctrlPts[3, 0, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
                Gl.glVertex3d(ctrlPts[0, 1, 0], ctrlPts[0, 1, 1], ctrlPts[0, 1, 2]);
                Gl.glVertex3d(ctrlPts[1, 1, 0], ctrlPts[1, 1, 1], ctrlPts[1, 1, 2]);
                Gl.glVertex3d(ctrlPts[2, 1, 0], ctrlPts[2, 1, 1], ctrlPts[2, 1, 2]);
                Gl.glVertex3d(ctrlPts[3, 1, 0], ctrlPts[3, 1, 1], ctrlPts[3, 1, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
                Gl.glVertex3d(ctrlPts[0, 2, 0], ctrlPts[0, 2, 1], ctrlPts[0, 2, 2]);
                Gl.glVertex3d(ctrlPts[1, 2, 0], ctrlPts[1, 2, 1], ctrlPts[1, 2, 2]);
                Gl.glVertex3d(ctrlPts[2, 2, 0], ctrlPts[2, 2, 1], ctrlPts[2, 2, 2]);
                Gl.glVertex3d(ctrlPts[3, 2, 0], ctrlPts[3, 2, 1], ctrlPts[3, 2, 2]);
            Gl.glEnd();
            Gl.glBegin(Gl.GL_LINE_STRIP);
                Gl.glVertex3d(ctrlPts[0, 3, 0], ctrlPts[0, 3, 1], ctrlPts[0, 3, 2]);
                Gl.glVertex3d(ctrlPts[1, 3, 0], ctrlPts[1, 3, 1], ctrlPts[1, 3, 2]);
                Gl.glVertex3d(ctrlPts[2, 3, 0], ctrlPts[2, 3, 1], ctrlPts[2, 3, 2]);
                Gl.glVertex3d(ctrlPts[3, 3, 0], ctrlPts[3, 3, 1], ctrlPts[3, 3, 2]);
            Gl.glEnd();
            Gl.glDisable(Gl.GL_LIGHTING);
        }        
        public void generateCurvePts()
        {
            double oneMinusU;
            double oneMinusV;
            double oneMinusUSqr;
            double oneMinusVSqr;
            double oneMinusUCube;
            double oneMinusVCube;
            double uSqr;
            double vSqr;
            double uCube;
            double vCube;
            double u, v;
            double[,] u2vPts = new double[4, 3];
            double x0, y0, z0;
            int m=0;
            int n=0;

            double inc1 = 1.0 / (NUM1 - 1);
            double lim1 = 1.0;
            double inc2 = 1.0 / (NUM2 - 1);
            double lim2 = 1.0;
            for (v = 0.0; v <= lim1 ; v += inc1)
            {
                
                n = 0;
                for (u = 0.0; u <= lim2; u += inc2)
                {
                    oneMinusU = 1 - u;
                    oneMinusV = 1 - v;
                    oneMinusUSqr = oneMinusU * oneMinusU;
                    oneMinusVSqr = oneMinusV * oneMinusV;
                    oneMinusUCube = oneMinusUSqr * oneMinusU;
                    oneMinusVCube = oneMinusVSqr * oneMinusV;
                    uSqr = u * u;
                    vSqr = v * v;
                    uCube = uSqr * u;
                    vCube = vSqr * v;

                    for (int i = 0; i < 4; i++)
                    {
                        x0 = oneMinusUCube * ctrlPts[i, 0, 0] +
                            3 * oneMinusUSqr * u * ctrlPts[i, 1, 0] +
                            3 * oneMinusU * uSqr * ctrlPts[i, 2, 0] +
                            uCube * ctrlPts[i, 3, 0];
                        u2vPts[i, 0] = x0;

                        y0 = oneMinusUCube * ctrlPts[i, 0, 1] +
                            3 * oneMinusUSqr * u * ctrlPts[i, 1, 1] +
                            3 * oneMinusU * uSqr * ctrlPts[i, 2, 1] +
                            uCube * ctrlPts[i, 3, 1];
                        u2vPts[i, 1] = y0;

                        z0 = oneMinusUCube * ctrlPts[i, 0, 2] +
                            3 * oneMinusUSqr * u * ctrlPts[i, 1, 2] +
                            3 * oneMinusU * uSqr * ctrlPts[i, 2, 2] +
                            uCube * ctrlPts[i, 3, 2];
                        u2vPts[i, 2] = z0;
                    }

                    x0 = oneMinusVCube * u2vPts[0, 0] +
                            3 * oneMinusVSqr * v * u2vPts[1, 0] +
                            3 * oneMinusV * vSqr * u2vPts[2, 0] +
                            vCube * u2vPts[3, 0];

                    y0 = oneMinusVCube * u2vPts[0, 1] +
                            3 * oneMinusVSqr * v * u2vPts[1, 1] +
                            3 * oneMinusV * vSqr * u2vPts[2, 1] +
                            vCube * u2vPts[3, 1];

                    z0 = oneMinusVCube * u2vPts[0, 2] +
                            3 * oneMinusVSqr * v * u2vPts[1, 2] +
                            3 * oneMinusV * vSqr * u2vPts[2, 2] +
                            vCube * u2vPts[3, 2];
                    curvePts[m, n, 0] = x0;
                    curvePts[m, n, 1] = y0;
                    curvePts[m, n, 2] = z0;
                    n++;
                }
                m++;
            }
        }
        public void generateNormals()
        {
            double[] vector1 = new double[3];
            double[] vector2 = new double[3];
            double mag;
            int m = 0, n = 0;
            for (m = 0; m <NUM1; m++)
            {
                for ( n= 0; n <NUM2; n++)
                {
                    vector1[0] = curvePts[m + 1, n, 0] - curvePts[m, n, 0];//
                    vector1[1] = curvePts[m + 1, n, 1] - curvePts[m, n, 1];
                    vector1[2] = curvePts[m + 1, n, 2] - curvePts[m, n, 2];

                    vector2[0] = curvePts[m, n + 1, 0] - curvePts[m, n, 0];
                    vector2[1] = curvePts[m, n + 1, 1] - curvePts[m, n, 1];
                    vector2[2] = curvePts[m, n + 1, 2] - curvePts[m, n, 2];

                    normals[m, n, 0] = vector2[1] * vector1[2] - vector1[1] * vector2[2];
                    normals[m, n, 1] = vector2[2] * vector1[0] - vector2[0] * vector1[2];
                    normals[m, n, 2] = vector2[0] * vector1[1] - vector1[0] * vector2[1];

                    mag = Math.Sqrt(normals[m, n, 0] * normals[m, n, 0] +
                                    normals[m, n, 1] * normals[m, n, 1] +
                                    normals[m, n, 2] * normals[m, n, 2]);
                    normals[m, n, 0] /= mag;
                    normals[m, n, 1] /= mag;
                    normals[m, n, 2] /= mag;
                }
            }
            for (n = 0; n < NUM2; n++)
            {
                normals[m, n, 0] = normals[m - 1, n, 0];
                normals[m, n, 1] = normals[m - 1, n, 0];
                normals[m, n, 2] = normals[m - 1, n, 0];
            }
        }

        private void meshButton_Click(object sender, EventArgs e)
        {
            drawFlag = 1;
            theta = 0;
            glControl.Invalidate();
        }
        private void surfaceButton_Click(object sender, EventArgs e)
        {
            drawFlag = 2;
            theta = 0;
            glControl.Invalidate();
        }
        private void rotateButton_Click(object sender, EventArgs e)
        {
            if (rotateButton.Text == "Rotate")
            {
                rotateFlag = 1;
                rotateButton.Text = "No Rotation";
            }
            else
            {
                rotateFlag = 0;
                rotateButton.Text = "Rotate";
            }
            glControl.Invalidate();
        }
        private void glControl_KeyDown(object sender, KeyEventArgs e)
        {
            if(e.KeyCode==Keys.Escape)
                Application.Exit();
        }
        #endregion
    }
}